# Copyright (C) 1998-2011 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""HyperArch: Pipermail archiving for Mailman

     - The Dragon De Monsyne <dragondm@integral.org>

"""

import os
import re
import sys
import gzip
import time
import errno
import urllib
import logging
import binascii
import mailbox

from tempfile import mkstemp
from shutil import move
from email.Charset import Charset
from email.Errors import HeaderParseError
from email.Header import decode_header, make_header
from flufl.lock import Lock, TimeOutError
from lazr.config import as_boolean
from string import Template
from zope.component import getUtility
from cStringIO import StringIO
from email.utils import parseaddr, parsedate_tz, mktime_tz, formatdate
from string import lowercase
from storm.locals import *

from archiver.model.article import StormArticle
from archiver.model.conversation import Conversation
from archiver.model.mlist import Mlist
from archiver.Indexer import Indexer

from archiver.core.i18n import _
#from mailman.config import config
#from mailman.core.i18n import _, ctime
#from mailman.interfaces.listmanager import IListManager
#from mailman.utilities.i18n import find
#from mailman.utilities.string import uncanonstr, websafe
SPACE = ' '
mbox_archive = 1
log = logging.getLogger('mailman.error')
log1= logging.getLogger('mailman.http')
EMPTYSTRING = ''
NL = '\n'


# MacOSX has a default stack size that is too small for deeply recursive
# regular expressions.  We see this as crashes in the Python test suite when
# running test_re.py and test_sre.py.  The fix is to set the stack limit to
# 2048; the general recommendation is to do in the shell before running the
# test suite.  But that's inconvenient for a daemon like the qrunner.
#
# AFAIK, this problem only affects the archiver, so we're adding this work
# around to this file (it'll get imported by the bundled pipermail or by the
# bin/arch script.  We also only do this on darwin, a.k.a. MacOSX.
if sys.platform == 'darwin':
    try:
        import resource
    except ImportError:
        pass
    else:
        soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
        newsoft = min(hard, max(soft, 1024*2048))
        resource.setrlimit(resource.RLIMIT_STACK, (newsoft, hard))

def prepare_url(subject):
    subject = re.sub("[^-A-Za-z0-9 ]","",subject)
    subject = re.sub("[ -]+","-",subject.strip().lower())
    return subject if len(subject) < 80 else subject[:80]

def safe_unicode(obj, *args):
    """ return the unicode representation of obj """
    try:
        return unicode(obj, *args)
    except UnicodeDecodeError:
        # obj is byte string
        ascii_text = str(obj).encode('string_escape')
        return unicode(ascii_text)

def html_quote(s, langcode=None):
    repls = ( ('&', '&amp;'),
              ("<", '&lt;'),
              (">", '&gt;'),
              ('"', '&quot;'))
    for thing, repl in repls:
        s = s.replace(thing, repl)
    return uncanonstr(s, langcode)


def url_quote(s):
    return urllib.quote(s)


def null_to_space(s):
    return s.replace('\000', ' ')


def sizeof(filename, lang):
    try:
        size = os.path.getsize(filename)
    except OSError, e:
        # ENOENT can happen if the .mbox file was moved away or deleted, and
        # an explicit mbox file name was given to bin/arch.
        if e.errno <> errno.ENOENT: raise
        return _('size not available')
    if size < 1000:
        with _.using(lang.code):
            out = _(' %(size)i bytes ')
        return out
    elif size < 1000000:
        return ' %d KB ' % (size / 1000)
    # GB?? :-)
    return ' %d MB ' % (size / 1000000)


html_charset = '<META http-equiv="Content-Type" ' \
               'content="text/html; charset=%s">'

def CGIescape(arg, lang=None):
    if isinstance(arg, unicode):
        s = websafe(arg)
    else:
        s = websafe(str(arg))
    return uncanonstr(s.replace('"', '&quot;'), lang.code)

# Parenthesized human name
paren_name_pat = re.compile(r'([(].*[)])')

# Subject lines preceded with 'Re:'
REpat = re.compile( r"\s*RE\s*(\[\d+\]\s*)?:\s*", re.IGNORECASE)

# E-mail addresses and URLs in text
emailpat = re.compile(r'([-+,.\w]+@[-+.\w]+)')

#  Argh!  This pattern is buggy, and will choke on URLs with GET parameters.
urlpat = re.compile(r'(\w+://[^>)\s]+)') # URLs in text

# Blank lines
blankpat = re.compile(r'^\s*$')

# Starting <html> directive
htmlpat = re.compile(r'^\s*<HTML>\s*$', re.IGNORECASE)
# Ending </html> directive
nohtmlpat = re.compile(r'^\s*</HTML>\s*$', re.IGNORECASE)
# Match quoted text
quotedpat = re.compile(r'^([>|:]|&gt;)+')



_templatefilepathcache = {}


msgid_pat = re.compile(r'(<.*>)')
def strip_separators(s):
    "Remove quotes or parenthesization from a Message-ID string"
    if not s:
        return ""
    if s[0] in '"<([' and s[-1] in '">)]':
        s = s[1:-1]
    return s

smallNameParts = ['van', 'von', 'der', 'de']

def fixAuthor(author):
    "Canonicalize a name into Last, First format"
    # If there's a comma, guess that it's already in "Last, First" format
    if ',' in author:
        return author
    L = author.split()
    i = len(L) - 1
    if i == 0:
        return author # The string's one word--forget it
    if author.upper() == author or author.lower() == author:
        # Damn, the name is all upper- or lower-case.
        while i > 0 and L[i-1].lower() in smallNameParts:
            i = i - 1
    else:
        # Mixed case; assume that small parts of the last name will be
        # in lowercase, and check them against the list.
        while i>0 and (L[i-1][0] in lowercase or
                       L[i-1].lower() in smallNameParts):
            i = i - 1
    author = SPACE.join(L[-1:] + L[i:-1]) + ', ' + SPACE.join(L[:i])
    return author

# The Article class encapsulates a single posting.  The attributes are:
#
#  sequence : Sequence number, unique for each article in a set of archives
#  subject  : Subject
#  datestr  : The posting date, in human-readable format
#  date     : The posting date, in purely numeric format
#  fromdate : The posting date, in `unixfrom' format
#  headers  : Any other headers of interest
#  author   : The author's name (and possibly organization)
#  email    : The author's e-mail address
#  msgid    : A unique message ID
#  in_reply_to : If !="", this is the msgid of the article being replied to
#  references: A (possibly empty) list of msgid's of earlier articles in
#              the thread
#  body     : A string making up the message body

class Article:
    _last_article_time = time.time()

    def __init__(self, message=None, sequence=0, keepHeaders=[],
                       lang=None, mlist=None):
        if message is None:
            return
        self.sequence = sequence

        self.parentID = None
        self.threadKey = None
        # otherwise the current sequence number is used.
        id = strip_separators(message['Message-Id'])
        if id == "":
            self.msgid = str(self.sequence)
        else: self.msgid = id

        if message.has_key('Subject'):
            self.subject = str(message['Subject'])
        else:
            self.subject = _('No subject')
        if self.subject == "": self.subject = _('No subject')

        self._set_date(message)

        # Figure out the e-mail address and poster's name.  Use the From:
        # field first, followed by Reply-To:
        self.author, self.email = parseaddr(message.get('From', ''))
        e = message['Reply-To']
        if not self.email and e is not None:
            ignoreauthor, self.email = parseaddr(e)
        self.email = strip_separators(self.email)
        self.author = strip_separators(self.author)

        if self.author == "":
            self.author = self.email

        # Save the In-Reply-To:, References:, and Message-ID: lines
        #
        # TBD: The original code does some munging on these fields, which
        # shouldn't be necessary, but changing this may break code.  For
        # safety, I save the original headers on different attributes for use
        # in writing the plain text periodic flat files.
        self._in_reply_to = message['in-reply-to']
        self._references = message['references']
        self._message_id = message['message-id']

        i_r_t = message['In-Reply-To']
        if i_r_t is None:
            self.in_reply_to = ''
            
        else:
            match = msgid_pat.search(i_r_t)
            if match is None: 
                self.in_reply_to = ''
            else: 
                self.in_reply_to = strip_separators(match.group(1))
        if not self.in_reply_to=='':
            self._in_reply_to = self.in_reply_to
            
        references = message['References']
        if references is None:
            self.references = []
        else:
            self.references = map(strip_separators, references.split())

        # Save any other interesting headers
        self.headers = {}
        for i in keepHeaders:
            if message.has_key(i):
                self.headers[i] = message[i]

        # Read the message body
        s = StringIO(message.get_payload(decode=True)\
                     or message.as_string().split('\n\n',1)[1])
#        self.body = s.readlines()
        self.body = s.read()

        self.prev = None
        self.next = None
        # Trim Re: from the subject line
        i = 0
        while i != -1:
            result = REpat.match(self.subject)
            if result:
                i = result.end(0)
                self.subject = self.subject[i:]
            else:
                i = -1
        # Useful to keep around
        self._lang = lang
        self._mlist = mlist
    
        # thread_addr is for dlists
        self.thread_addr = ''
        addrs_to_try = ['To', 'X-Original-To', 'Delivered-To', 'Cc', 'Bcc']
        i = 0
        while (self.thread_addr == None or self.thread_addr.find('+') == -1 or self.thread_addr.find(',') != -1 or self.thread_addr.find(' ') != -1) and i < len(addrs_to_try):
            self.thread_addr = message[addrs_to_try[i]]
            i += 1
            if self.thread_addr != None:
                self.thread_addr = self.thread_addr.replace('>', '').replace('<', '')
            else:
                self.thread_addr = self.strip_subject(self.subject)
                
#        if as_boolean(config.archiver.pipermail.obscure_email_addresses):
#            # Avoid i18n side-effects.  Note that the language for this
#            # article (for this list) could be different from the site-wide
#            # preferred language, so we need to ensure no side-effects will
#            # occur.  Think what happens when executing bin/arch.
#            with _.using(lang.code):
#                if self.author == self.email:
#                    self.author = self.email = re.sub('@', _(' at '),
#                                                      self.email)
#                else:
#                    self.email = re.sub('@', _(' at '), self.email)
        # Snag the content-* headers.  RFC 1521 states that their values are
        # case insensitive.
        ctype = message.get('Content-Type', 'text/plain')
        cenc = message.get('Content-Transfer-Encoding', '')
        self.ctype = ctype.lower()
        self.cenc = cenc.lower()
        self.decoded = {}
        cset = 'us-ascii' #mlist.preferred_language.charset
        cset_out = Charset(cset).output_charset or cset
        charset = message.get_content_charset(cset_out)
        if charset:
            charset = charset.lower().strip()
            if charset[0]=='"' and charset[-1]=='"':
                charset = charset[1:-1]
            if charset[0]=="'" and charset[-1]=="'":
                charset = charset[1:-1]
            try:
                body = message.get_payload(decode=True)
            except binascii.Error:
                body = None
            if body and charset != 'us-ascii':
                # decode body
                try:
                    body = unicode(body, charset)
                except (UnicodeError, LookupError):
                    body = None
            if body:
#                self.body = [l + "\n" for l in body.splitlines()]
                self.body = body

        self.decode_headers()

    def quote(self, buf):
        return html_quote(buf, self._lang.code)

    def decode_headers(self):
        """MIME-decode headers.

        If the email, subject, or author attributes contain non-ASCII
        characters using the encoded-word syntax of RFC 2047, decoded versions
        of those attributes are placed in the self.decoded (a dictionary).

        If the list's charset differs from the header charset, an attempt is
        made to decode the headers as Unicode.  If that fails, they are left
        undecoded.
        """
        author = self.decode_charset(self.author)
        subject = self.decode_charset(self.subject)
        if author:
            self.decoded['author'] = author
            email = self.decode_charset(self.email)
            if email:
                self.decoded['email'] = email
        if subject:
#            if as_boolean(config.archiver.pipermail.obscure_email_addresses):
#                with _.using(self._lang.code):
#                    atmark = _(' at ')
#                    subject = re.sub(r'([-+,.\w]+)@([-+.\w]+)',
#                              '\g<1>' + atmark + '\g<2>', subject)
            self.decoded['subject'] = subject
        self.decoded['stripped'] = self.strip_subject(subject or self.subject)

    def strip_subject(self, subject):
        # Strip subject_prefix and Re: for subject sorting
        # This part was taken from CookHeaders.py (TK)
        prefix = None #self._mlist.subject_prefix.strip()
        if prefix:
            prefix_pat = re.escape(prefix)
            prefix_pat = '%'.join(prefix_pat.split(r'\%'))
            prefix_pat = re.sub(r'%\d*d', r'\s*\d+\s*', prefix_pat)
            subject = re.sub(prefix_pat, '', subject)
        subject = subject.lstrip()
        strip_pat = re.compile('^((RE|AW|SV|VS)(\[\d+\])?:\s*)+', re.I)
        stripped = strip_pat.sub('', subject)
        return stripped

    def decode_charset(self, field):
        # TK: This function was rewritten for unifying to Unicode.
        # Convert 'field' into Unicode one line string.
        try:
            pairs = decode_header(field)
            ustr = make_header(pairs).__unicode__()
        except (LookupError, UnicodeError, ValueError, HeaderParseError):
            # assume list's language
            cset = 'us-ascii' #self._mlist.preferred_language.charset
            if cset == 'us-ascii':
                cset = 'iso-8859-1' # assume this for English list
            ustr = unicode(field, cset, 'replace')
        return u''.join(ustr.splitlines())

    def _get_subject_enc(self, art):
        """Return the subject of art, decoded if possible.

        If the charset of the current message and art match and the
        article's subject is encoded, decode it.
        """
        return art.decoded.get('subject', art.subject)

    _rx_quote = re.compile('=([A-F0-9][A-F0-9])')
    _rx_softline = re.compile('=[ \t]*$')

    def _get_body(self):
        """Return the message body ready for HTML, decoded if necessary"""
        body = self.body
        return null_to_space(EMPTYSTRING.join(body))

    def _set_date(self, message):
        def floatdate(header):
            missing = []
            datestr = message.get(header, missing)
            if datestr is missing:
                return None
            date = parsedate_tz(datestr)
            try:
                return mktime_tz(date)
            except (TypeError, ValueError, OverflowError):
                return None
        date = floatdate('date')
        if date is None:
            date = floatdate('x-list-received-date')
        if date is None:
            # What's left to try?
            date = self._last_article_time + 1
        self._last_article_time = date
        self.date = date
#        self.date = '%011i' % date
        print type(self.date)
        self.datestr = message.get('date') \
                       or message.get('x-list-received-date') \
                       or formatdate(date)

        self.fromdate = time.ctime(int(self.date))
 
class T:
    DIRMODE = 0755      # Mode to give to created directories

    def __init__(self, basedir = None, mlist = None):
        # If basedir isn't provided, assume the current directory
        self.mlist = mlist
        if basedir is None:
            basedir = os.getcwd()
        else:
            basedir = os.path.expanduser(basedir)
        self.listdir = os.path.join(basedir,mlist)
        # If the directory doesn't exist, create it.
        for dir in (basedir, self.listdir):
            try:
                os.stat(dir)
            except os.error, errdata:
                errno, errmsg = errdata
                if errno != 2:
                    raise os.error, errdata
                else:
                    omask = os.umask(0)
                    try:
                        os.mkdir(dir, self.DIRMODE)
                    finally:
                        os.umask(omask)
        #find db_path
        db_path = os.path.join(self.listdir,'archives.db')
        self.store = Store(create_database('sqlite:///'+db_path))
        
        #insert mlist into mlist db.
        mlist_db= os.path.join(basedir,'mlists.db')
        store = Store(create_database('sqlite:///'+mlist_db))
        try:
            store.execute("CREATE TABLE mlist " "(list_name VARCHAR(255), db_path VARCHAR(255), id INTEGER PRIMARY KEY)")
        except:
            pass
        if store.find(Mlist, Mlist.list_name==unicode(self.mlist)).is_empty():
            store.add(Mlist(self.mlist,self.listdir))
            store.commit()
        
        #find index_path
        index_path = os.path.join(self.listdir,'archives.index')
        self.indexer = Indexer(index_path)
                
    def db_init(self):
        try:
            self.store.execute("CREATE TABLE article " "(subject VARCHAR(255), date FLOAT, datestr VARCHAR(255), author VARCHAR(255), email VARCHAR(255), msgid VARCHAR(255), id INTEGER PRIMARY KEY, in_reply_to VARCHAR(255), body TEXT, thread_addr VARCHAR(255), threadkey VARCHAR(255), pdate_msg INTEGER, ndate_msg INTEGER, pthread_msg INTEGER, nthread_msg INTEGER )")
        except:
            pass
        try:
            self.store.execute("CREATE TABLE conversation" "(subject VARCHAR(255), date FLOAT, datestr VARCHAR(255), thread_addr VARCHAR(255), author VARCHAR(255), count INTEGER, pdate_con VARCHAR(255), ndate_con VARCHAR(255))")
        except:
            pass

    def _makeArticle(self, msg):
        return Article(msg)

    def mboxToDB(self, path, start=None, end=None):
        mbox = iter(mailbox.mbox(path))
        if start is None:
            start = 0
        counter = 0
        while counter < start:
            try:
                m = next(mbox)
            except errors.DiscardMessage:
                continue
            if m is None:
                return
            counter += 1
        while True:
            try:
                m = next(mbox)
            except StopIteration:
                break
            except errors.DiscardMessage:
                continue
            except Exception:
                log.error('uncaught archiver exception')
                raise
            if m == '':
                # It was an unparseable message
                continue
#            self.message(_('#%(counter)05d %(msgid)s'))
            stormarticle = self.archiveMsg(m, commit = False)
            self.indexer.index_article(stormarticle,created= True, commit = False)
            if end is not None and counter >= end:
               break
            counter += 1
        #update navigation fields
        self.update_cons_date()
        self.update_msgs_thread()
        
        #commit database and index
        self.store.commit()
        self.indexer.commit()
   
    def archiveMsg(self, msg, commit=True):
        article = self._makeArticle(msg)
        if article.decoded.has_key('author'):
            author = fixAuthor(article.decoded['author'])
        else:
            author = fixAuthor(article.author)
        if article.decoded.has_key('stripped'):
            subject = article.decoded['stripped'].lower()
        else:
            subject = article.subject.lower()
        
        parent, in_reply = self.get_parent(article)
        
        #if article's parent exists
        if parent is not None:
            article.thread_addr = parent.thread_addr
            article.threadkey = parent.threadkey + str(article.date) + '.'
            #update corresponding conversation
            conversation  = self.store.get(Conversation, parent.thread_addr)
            if conversation is not None:
                conversation.count = conversation.count +1
                conversation.date = article.date
        else:
            article.thread_addr = unicode(prepare_url(article.thread_addr))
            article.threadkey =  str(article.date) + '.'
            #create new conversation object
            conversation = self.store.add(Conversation(article.thread_addr, float(article.date), unicode(article.subject), unicode(author), unicode(article.datestr), 1))
        
        id = None if in_reply is None else in_reply.id
        
        #update previous and next article sorted by date
        pdate_article = self.store.find(StormArticle).order_by(Desc(StormArticle.date)).first()
        
        stormarticle = StormArticle(unicode(subject), float(article.date), unicode (article.datestr), unicode(author), unicode(article.email), unicode(article.msgid), id, safe_unicode(article.body), article.thread_addr, unicode(article.threadkey))
        self.store.add(stormarticle)
        self.store.flush()
        
        if pdate_article is not None:
            pdate_article.ndate_msg = stormarticle.id
            stormarticle.pdate_msg = pdate_article.id 
        
        if commit:
            self.store.commit()
        return stormarticle
        
    def get_parent(self, article):
        parentID = None
        in_reply = None
        if article.in_reply_to:
            parent = self.store.find(StormArticle, StormArticle.msgid==unicode(article.in_reply_to))
            if parent is not None:
                parentID = parent.any()
                in_reply = parentID 
        elif article.references:
            for ref in article.references:
                parent = self.store.find(StormArticle, StormArticle.msgid==unicode(ref))
                if parent is not None:
                    parentID = parent.any()
                if parentID:
                    break
        return parentID, in_reply
    
    def update_msgs(self):
        # for now, assume id contains the thread_addr
        result = self.store.find(StormArticle)
        if result.is_empty():
            return
        messages = result.order_by(Asc(StormArticle.date))
        for i in range(messages.count()):
            messages[i].pdate_msg = messages[i-1].id
            try:
                messages[i].ndate_msg = messages[i+1].id
            except IndexError:
                pass
        self.store.commit()
        return 'Finished updating msgs'
     
    def update_cons_date(self):
        result = self.store.find(Conversation)
        if result.is_empty():
            raise EmptyTable("Conversation table is empty")
        conversations = result.order_by(Asc(Conversation.date))
        for i in range(conversations.count()):
            conversations[i].pdate_con = conversations[i-1].thread_addr
            try:
                conversations[i].ndate_con = conversations[i+1].thread_addr
            except IndexError:
                pass
        self.store.commit()
    
    def update_msgs_thread(self):
        result = self.store.find(StormArticle)
        if result.is_empty():
            raise EmptyTable("StormArticle table is empty")
        messages = result.order_by(Asc(StormArticle.threadkey))
        for i in range(messages.count()):
            messages[i].pthread_msg = messages[i-1].id
            try:
                messages[i].nthread_msg = messages[i+1].id
            except IndexError:
                pass
        self.store.commit()
    
    def update_cons(self, con):
        p_con = self.store.get(Conversation, con.pdate_con)
        n_con = self.store.get(Conversation, con.ndate_con)
        
        p_con.ndate_con = n_con.thread_addr
        n_con.pdate_con = p_con.thread_addr
        
        last_con = self.store.find(Conversation).order_by(Desc(Conversation.date)).first()
        last_con.ndate_con = con.thread_addr
        con.pdate_con = last_con.thread_addr