# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


# Python imports
import os
import sys

# Whoosh import
from whoosh.spelling import SpellChecker
from whoosh import store, fields, index
from whoosh.analysis import StemmingAnalyzer
from whoosh.filedb.filestore import FileStorage

# Pylons import
from archiver.model.article import StormArticle
from storm.locals import *

WHOOSH_INDEX = 'archives.index'
WHOOSH_SCHEMA = fields.Schema(msgid=fields.NUMERIC(stored=True, unique=True),
                                author=fields.TEXT(),
                                subject=fields.TEXT(analyzer=StemmingAnalyzer()),
                                body=fields.TEXT(analyzer=StemmingAnalyzer()),
                            )

class Indexer:
    '''This class is a utility class to quickly create or update an index.'''

    index = None

    writer = None

    storage = None

    last_index_date = None

    def __init__(self, indexpath):
        '''Accepts a path and initializes the index. If:
            - The path doesn't exist, it creates the path.
            - If the path is empty, it initializes an index.
            - If the path is full, it can be used to update the index.
        '''
        
        if os.path.exists(indexpath):
            if index.exists_in(indexpath):
                self.index = index.open_dir(indexpath)
            else:
                self.index = index.create_in(indexpath, schema = WHOOSH_SCHEMA)
        else:
            os.mkdir(indexpath)
            self.index = index.create_in(indexpath, schema = WHOOSH_SCHEMA)

        self.writer = self.index.writer()

        self.storage = self.index.storage

    def index_article(self, article, created, commit = True):
        if created:
            self.writer.add_document(msgid = article.id, author = article.author,
                                 subject = article.subject, body = article.body)
        else:
            #else just update the document index
            self.writer.add_document(msgid = article.id, author = article.author,
                                 subject = article.subject, body = article.body)
        if commit:
            self.writer.commit()
    def commit(self):
        self.writer.commit()

    def index_archives(self, store):
#        store = request.environ['storm.store'][mlist]
#        if self.last_index_date is not None:
#            articles = store.find(StormArticle, StormArticle.date > self.last_index_date)
#        else:
        articles = store.find(StormArticle)
        for article in articles:
#            if self.last_index_date < article.date:
#                self.last_index_date = article.date  
#            self.index_article(article, True)
            self.writer.add_document(msgid = article.id, author = article.author,
                                 subject = article.subject, body = article.body)
        self.writer.commit()

    def close(self):
        '''Close the indexer and create a spelling dictionary'''

        self.writer.commit()

        # TODO Delete previous files and remake the dictionary on every commit.
        # I know this is not optimal but Whoosh currently does not have any way
        # simply update a dictionary with a new document (like the index).
#        speller = SpellChecker(self.storage)
#
#        for k in IndexSchema.getspelledfields():
#            speller.add_field(self.index, k)
